@@ -28,30 +28,10 @@
 #endif
 #include <stdio.h>
 
-/* Compatibility with python 2.4 regarding python size type (PEP 353) */
-#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
-typedef int Py_ssize_t;
-#define PY_SSIZE_T_MAX INT_MAX
-#define PY_SSIZE_T_MIN INT_MIN
-#endif
-
-#if PY_MAJOR_VERSION >= 3
-#define IS_PY3K
-#define BYTES_CHAR "y"
-#define BYTES_TUPLE "yy#"
-#else
-#define BYTES_CHAR "s"
-#define BYTES_TUPLE "ss#"
-#define PyBytes_Check PyString_Check
-#define PyBytes_AS_STRING PyString_AS_STRING
-#define PyBytes_FromStringAndSize PyString_FromStringAndSize
-#define PyBytes_FromString PyString_FromString
-#endif
-
 #define ITEM_DOC \
-    ":param item: a string representing a file-name, or a file-like\n" \
-    "    object, or a file descriptor; this represents the file on \n" \
-    "    which to act\n"
+    ":param item: a string representing a file-name, a file-like\n" \
+    "    object, a file descriptor, or (in Python 3.6+) a path-like\n" \
+    "    object; this represents the file on which to act\n"
 
 #define NOFOLLOW_DOC \
     ":param nofollow: if true and if\n" \
@@ -150,33 +130,28 @@ static int merge_ns(const char *ns, const char *name,
 static int convert_obj(PyObject *myobj, target_t *tgt, int nofollow) {
     int fd;
     tgt->tmp = NULL;
-    if(PyBytes_Check(myobj)) {
-        tgt->type = nofollow ? T_LINK : T_PATH;
-        tgt->name = PyBytes_AS_STRING(myobj);
-    } else if(PyUnicode_Check(myobj)) {
-        tgt->type = nofollow ? T_LINK : T_PATH;
-        tgt->tmp = \
-            PyUnicode_AsEncodedString(myobj,
-                                      Py_FileSystemDefaultEncoding,
-#ifdef IS_PY3K
-                                      "surrogateescape"
-#else
-                                      "strict"
-#endif
-                                      );
-        if(tgt->tmp == NULL)
-            return -1;
-        tgt->name = PyBytes_AS_STRING(tgt->tmp);
-    } else if((fd = PyObject_AsFileDescriptor(myobj)) != -1) {
+    if((fd = PyObject_AsFileDescriptor(myobj)) != -1) {
         tgt->type = T_FD;
         tgt->fd = fd;
+        return 0;
+    }
+    // PyObject_AsFileDescriptor sets an error when failing, so clear
+    // it such that further code works; some method lookups fail if an
+    // error already occured when called, which breaks at least
+    // PyOS_FSPath (called by FSConverter).
+    PyErr_Clear();
+
+    if(PyUnicode_FSConverter(myobj, &(tgt->tmp))) {
+        tgt->type = nofollow ? T_LINK : T_PATH;
+        tgt->name = PyBytes_AS_STRING(tgt->tmp);
+        return 0;
     } else {
-        PyErr_SetString(PyExc_TypeError, "argument must be string or int");
+        // Don't set our own exception type, since we'd ignore the
+        // FSConverter-generated one.
         tgt->type = T_PATH;
         tgt->name = NULL;
         return -1;
     }
-    return 0;
 }
 
 /* Combine a namespace string and an attribute name into a
@@ -537,7 +512,7 @@ xattr_get(PyObject *self, PyObject *args, PyObject *keywds)
     static char *kwlist[] = {"item", "name", "nofollow", "namespace", NULL};
 
     /* Parse the arguments */
-    if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oet|i" BYTES_CHAR, kwlist,
+    if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oet|iy", kwlist,
                                      &myarg, NULL, &attrname, &nofollow, &ns))
         return NULL;
     res = NULL;
@@ -624,7 +599,7 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
     int io_errno;
 
     /* Parse the arguments */
-    if (!PyArg_ParseTupleAndKeywords(args, keywds, "O|i" BYTES_CHAR, kwlist,
+    if (!PyArg_ParseTupleAndKeywords(args, keywds, "O|iy", kwlist,
                                      &myarg, &nofollow, &ns))
         return NULL;
     if(convert_obj(myarg, &tgt, nofollow) < 0)
@@ -665,12 +640,15 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
             goto free_buf_val;
           }
         }
-        my_tuple = Py_BuildValue(BYTES_TUPLE, name, buf_val, nval);
+        my_tuple = Py_BuildValue("yy#", name, buf_val, nval);
         if (my_tuple == NULL) {
           Py_DECREF(mylist);
           goto free_buf_val;
         }
-        PyList_Append(mylist, my_tuple);
+        if(PyList_Append(mylist, my_tuple) < 0) {
+            Py_DECREF(mylist);
+            goto free_buf_val;
+        }
         Py_DECREF(my_tuple);
     }
 
@@ -807,7 +785,7 @@ xattr_set(PyObject *self, PyObject *args, PyObject *keywds)
                              "nofollow", "namespace", NULL};
 
     /* Parse the arguments */
-    if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oetet#|ii" BYTES_CHAR,
+    if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oetet#|iiy",
                                      kwlist, &myarg, NULL, &attrname, NULL,
                                      &buf, &bufsize_s, &flags, &nofollow, &ns))
         return NULL;
@@ -938,7 +916,7 @@ xattr_remove(PyObject *self, PyObject *args, PyObject *keywds)
     static char *kwlist[] = {"item", "name", "nofollow", "namespace", NULL};
 
     /* Parse the arguments */
-    if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oet|i" BYTES_CHAR, kwlist,
+    if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oet|iy", kwlist,
                                      &myarg, NULL, &attrname, &nofollow, &ns))
         return NULL;
 
@@ -1085,7 +1063,7 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
     static char *kwlist[] = {"item", "nofollow", "namespace", NULL};
 
     /* Parse the arguments */
-    if (!PyArg_ParseTupleAndKeywords(args, keywds, "O|i" BYTES_CHAR, kwlist,
+    if (!PyArg_ParseTupleAndKeywords(args, keywds, "O|iy", kwlist,
                                      &myarg, &nofollow, &ns))
         return NULL;
     res = NULL;
@@ -1184,7 +1162,7 @@ static char __xattr_doc__[] = \
     "   a :exc:`EnvironmentError`; under\n"
     "   Linux, the following ``errno`` values are used:\n"
     "\n"
-    "   - ``ENODATA`` means that the attribute name is\n invalid\n"
+    "   - ``ENODATA`` means that the attribute name is invalid\n"
     "   - ``ENOTSUP`` and ``EOPNOTSUPP`` mean that the filesystem does not\n"
     "     support extended attributes, or that the namespace is invalid\n"
     "   - ``E2BIG`` mean that the attribute value is too big\n"
@@ -1197,8 +1175,6 @@ static char __xattr_doc__[] = \
     "\n"
     ;
 
-#ifdef IS_PY3K
-
 static struct PyModuleDef xattrmodule = {
     PyModuleDef_HEAD_INIT,
     "xattr",
@@ -1212,23 +1188,14 @@ static struct PyModuleDef xattrmodule = {
 PyMODINIT_FUNC
 PyInit_xattr(void)
 
-#else
-#define INITERROR return
-void
-initxattr(void)
-#endif
 {
     PyObject *ns_security = NULL;
     PyObject *ns_system   = NULL;
     PyObject *ns_trusted  = NULL;
     PyObject *ns_user     = NULL;
-#ifdef IS_PY3K
     PyObject *m = PyModule_Create(&xattrmodule);
-#else
-    PyObject *m = Py_InitModule3("xattr", xattr_methods, __xattr_doc__);
-#endif
     if (m==NULL)
-        INITERROR;
+        return NULL;
 
     PyModule_AddStringConstant(m, "__author__", _XATTR_AUTHOR);
     PyModule_AddStringConstant(m, "__contact__", _XATTR_EMAIL);
@@ -1262,11 +1229,7 @@ initxattr(void)
         goto err_out;
     ns_user = NULL;
 
-#ifdef IS_PY3K
     return m;
-#else
-    return;
-#endif
 
  err_out:
     Py_XDECREF(ns_user);
